# -*- python -*-
# $RCSfile: skeletonbdydisplay.py,v $
# $Revision: 1.25 $
# $Author: langer $
# $Date: 2007/03/13 20:50:25 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from oof2.common import color
from oof2.common import debug
from oof2.common import primitives
from oof2.common import registeredclass
from oof2.common.IO import display
from oof2.common.IO import ghostgfxwindow
from oof2.common.IO import gfxmanager
from oof2.common.IO import mainmenu
from oof2.common.IO import oofmenu
from oof2.common.IO import parameter
from oof2.common.IO import xmlmenudump
from oof2.engine import skeletonboundary
import math
import string

# Special parameter class for skeleton boundaries.  Returns a
# list of strings corresponding to the names of boundaries in a
# given skeleton.  The custom class is primarily to allow for a
# special skeleton-smart widget.

class SkeletonBoundaryListParameter(parameter.ListOfStringsParameter):
    pass

# Layer for showing the boundaries (point and edge) of a skeleton.
class SkeletonBoundaryDisplay(display.DisplayMethod):
    def __init__(self, boundaries, color, linewidth, dotsize, arrowsize):
        self.boundaries = boundaries
        self.color = color
        self.linewidth = linewidth
        self.dotsize = dotsize
        self.arrowsize = arrowsize
        display.DisplayMethod.__init__(self)

    def draw(self, gfxwindow, device):
        skel = self.who().resolve(gfxwindow)
        skelobj = skel.getObject()
        device.set_lineColor(self.color)
        device.set_lineWidth(self.linewidth)
        for k in self.boundaries:
            try:
                b = skelobj.edgeboundaries[k]
            except KeyError:
                pass
            else:
                for e in b.edges:
                    nodes = e.get_nodes()
                    device.set_lineWidth(self.linewidth)
                    device.draw_segment(primitives.Segment(
                        nodes[0].position(), nodes[1].position()))
                    # Boundaries are directed from 0 to 1.
                    center = (nodes[0].position() + nodes[1].position())/2
                    diff = (nodes[1].position() - nodes[0].position())
                    # Zero of angles is the y-axis, not the x-axis...
                    angle = math.atan2(-diff.x, diff.y)
                    device.set_lineWidth(self.arrowsize)
                    device.draw_triangle(center, angle)
            
        device.set_lineWidth(self.dotsize)
        device.set_fillColor(self.color)
        for k in self.boundaries:
            try:
                b = skelobj.pointboundaries[k]
            except KeyError:
                pass
            else:
                for n in b.nodes:
                    device.draw_dot(n.position())

    def getTimeStamp(self, gfxwindow):
        return max( self.timestamp,
                    self.who().resolve(gfxwindow).bdytimestamp )

    # Need to override hash, because we contain a list.  For this
    # class, object identity is a good test of equality.

    ## TODO: Using id(self) for hashing is not recommended in Python
    ## 2.5, since ids can be longs and hash must be an int.  Why do we
    ## really need __hash__ here?  Is this used as a key?
##    def __hash__(self):
##        return id(self)
    
                    
skeletonBoundaryDisplay = registeredclass.Registration(
    'Boundaries', display.DisplayMethod,
    SkeletonBoundaryDisplay,
    params=[SkeletonBoundaryListParameter('boundaries', [],
                                      tip="Boundaries to display."),
            color.ColorParameter('color', value=color.gray50,
                                 tip="Color for the displayed boundaries."),
            parameter.IntRangeParameter('linewidth', (0, 10), 4,
                                        tip="Line width for edge boundaries."),
            parameter.IntRangeParameter('dotsize', (0, 10), 4,
                                        tip="Dot radius for point boundaries."),
            parameter.IntRangeParameter('arrowsize', (0, 20), 10,
                                        tip="Arrow size for edge boundaries.")],
    ordering=1.0,
    layerordering=display.SemiLinear(2),
    whoclasses=('Skeleton',),
    tip="Display some or all of the boundaries of the Skeleton",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/skeletonbdydisplay.xml')
    )

# Layer for showing the selected boundary (point and edge) of a skeleton.

class SelectedSkeletonBoundaryDisplay(display.DisplayMethod):
    def __init__(self, color, linewidth, dotsize, arrowsize):
        self.color = color
        self.linewidth = linewidth
        self.dotsize = dotsize
        self.arrowsize = arrowsize
        display.DisplayMethod.__init__(self)
        
    def draw(self, gfxwindow, device):
        skel = self.who().resolve(gfxwindow)
        skelobj = skel.getObject()
        bdy = skel.getSelectedBoundary()  # SkelContextBoundary
        if bdy is not None:
            # bdy.draw calls either drawEdgeBoundary or drawPointBoundary
            bdy.draw(self, device, skelobj)
    def drawEdgeBoundary(self, bdy, skelobj, device):
        b = bdy.boundary(skelobj)
        device.set_lineColor(self.color)
        for e in b.edges:
            nodes = e.get_nodes()
            device.set_lineWidth(self.linewidth)
            n0 = nodes[0].position()
            n1 = nodes[1].position()
            device.draw_segment(primitives.Segment(n0, n1))
            # Boundaries are directed from 0 to 1.
            center = (n0 + n1)/2
            diff = (n1 - n0)
            # Zero of angles is the y-axis, not the x-axis...
            angle = math.atan2(-diff.x, diff.y)
            device.set_lineWidth(self.arrowsize)
            device.draw_triangle(center, angle)
    def drawPointBoundary(self, bdy, skelobj, device):
        b = bdy.boundary(skelobj)
        device.set_lineWidth(self.dotsize)
        device.set_lineColor(self.color)
        for n in b.nodes:
            device.draw_dot(n.position())
    
    def getTimeStamp(self, gfxwindow):
        skelcontext = self.who().resolve(gfxwindow)
        bdy = skelcontext.getSelectedBoundary()
        if bdy is not None:
            return max(self.timestamp, skelcontext.bdyselected,
                       skelcontext.getSelectedBoundary().timestamp)
        return max(self.timestamp, skelcontext.bdyselected)

defaultSelSkelBdyColor = color.orange
defaultSelSkelBdyLineWidth = 4
defaultSelSkelBdyDotSize = 4
defaultSelSkelBdyArrowSize = 10

def _setSelSkelBdyParams(menuitem, color, linewidth, dotsize, arrowsize):
    global defaultSelSkelBdyColor
    global defaultSelSkelBdyLineWidth
    global defaultSelSkelBdyDotSize
    global defaultSelSkelBdyArrowSize
    defaultSelSkelBdyColor = color
    defaultSelSkelBdyLineWidth = linewidth
    defaultSelSkelBdyDotSize = dotsize
    defaultSelSkelBdyArrowSize = arrowsize

selskelbdyparams = [
    color.ColorParameter('color', value=defaultSelSkelBdyColor,
                         tip="Color for the selected boundary."),
    parameter.IntRangeParameter('linewidth', (0, 10),
                                defaultSelSkelBdyLineWidth,
                                tip="Line width for edge boundaries."),
    parameter.IntRangeParameter('dotsize', (0, 10),
                                defaultSelSkelBdyDotSize,
                                tip="Dot radius for point boundaries."),
    parameter.IntRangeParameter('arrowsize', (0, 20),
                                defaultSelSkelBdyArrowSize,
                                tip="Arrow size for edge boundaries.")]

mainmenu.gfxdefaultsmenu.Skeletons.addItem(oofmenu.OOFMenuItem(
    "Selected_Boundary",
    callback=_setSelSkelBdyParams,
    params=selskelbdyparams,
    ordering=4.5,
    help="Set default parameters for displaying selected skeleton boundaries.",
    discussion="""<para>

    This command sets the default parameters for the
    <xref linkend="RegisteredClass:SelectedSkeletonBoundaryDisplay"/>,
    which displays the currently
    <link linkend="Section:Tasks:SkeletonBoundaries">selected
    <classname>Skeleton</classname> boundaries</link>.  Put this command
    in the &oof2rc; file to set default values for all &oof2; sessions.
    
    </para>"""))
                    
selectedSkeletonBoundaryDisplay = registeredclass.Registration(
    'Selected Boundary', display.DisplayMethod,
    SelectedSkeletonBoundaryDisplay,
    params=selskelbdyparams,
    ordering=2.3,
    layerordering=display.SemiLinear(1),
    whoclasses=('Skeleton',),
    tip="Display the currently selected boundary.",
    discussion = xmlmenudump.loadFile('DISCUSSIONS/engine/reg/skeletonselbdydisplay.xml')
    )

def defaultSelectedSkeletonBoundaryDisplay():
    return selectedSkeletonBoundaryDisplay(
        color=defaultSelSkelBdyColor,
        linewidth=defaultSelSkelBdyLineWidth,
        dotsize=defaultSelSkelBdyDotSize,
        arrowsize=defaultSelSkelBdyArrowSize)

ghostgfxwindow.PredefinedLayer('Skeleton', '<topmost>',
                               defaultSelectedSkeletonBoundaryDisplay)
