# -*- python -*-
# $RCSfile: autoskeleton.py,v $
# $Revision: 1.4 $
# $Author: langer $
# $Date: 2007/03/13 20:50:25 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

## This file defines a menu item that aggregates a bunch of other menu
## items in an attempt to create and adapt a Skeleton completely
## automatically.  The user must provide two length scales -- the size
## of the largest and smallest features of interest in the
## microstructure.  An initial Skeleton is created at the larger
## length scale and refined by homogeneity (using the provided
## threshold) to the smaller scale.  Then SnapRefine is applied once,
## the Skeleton is Rationalized, and finally it's Smoothed (after
## pinning internal boundary nodes).

from oof2.common import debug
from oof2.common import enum
from oof2.common import utils
from oof2.common.IO import automatic
from oof2.common.IO import oofmenu
from oof2.common.IO import parameter
from oof2.common.IO import whoville
from oof2.common.IO import xmlmenudump
from oof2.engine.IO import skeletonmenu
import oof2.common.microstructure
import oof2.common.units
import math
import sys

def _autoSkeleton(menuitem, name, microstructure,
                  left_right_periodicity, top_bottom_periodicity,
                  maxscale, minscale, units, threshold):
    # Run the actual callback in the main namespace, so that it can
    # use menu and registeredclass names trivially.
    utils.OOFrun(_autoSkeletonMain, name, microstructure,
                 left_right_periodicity, top_bottom_periodicity,
                 maxscale, minscale,
                 units, threshold)
    
def _autoSkeletonMain(name, microstructure,
                      left_right_periodicity,
                      top_bottom_periodicity,
                      maxscale, minscale, units,
                      threshold):
    ms = oof2.common.microstructure.microStructures[microstructure]
    if units == 'Pixel':
        size = ms.getObject().sizeInPixels()
        nx0, ny0 = size/maxscale
        nx1, ny1 = size/minscale
    else:                               # physical units
        size = ms.getObject().size()
        nx0, ny0 = size/maxscale
        nx1, ny1 = size/minscale

    nx0 = max(nx0, 1)
    ny0 = max(ny0, 1)
    
    nrefine_x = math.log(nx1/nx0)/math.log(2)
    nrefine_y = math.log(ny1/ny0)/math.log(2)
    nrefine = int(math.ceil(max(nrefine_x, nrefine_y)))

    skelname = microstructure+":"+name

    OOF.Skeleton.New(name=name, 
                     microstructure=microstructure,
                     x_elements=int(math.ceil(nx0)),
                     y_elements=int(math.ceil(ny0)),
                     skeleton_geometry=QuadSkeleton(
        left_right_periodicity=left_right_periodicity,
        top_bottom_periodicity=top_bottom_periodicity))

    for i in range(nrefine):
        OOF.Skeleton.Modify(
            skeleton=skelname,
            modifier=Refine(targets=CheckHomogeneity(threshold=threshold),
                            criterion=Unconditionally(),
                            degree=Bisection(rule_set='liberal'),
                            alpha=0.8))
                     
    OOF.Skeleton.Modify(
        skeleton=skelname,
        modifier=SnapRefine(targets=CheckHomogeneity(threshold=threshold),
                            criterion=Unconditionally(),
                            min_distance=0.0))
    for i in range(2):
        OOF.Skeleton.Modify(
            skeleton=skelname,
            modifier=Rationalize(targets=AllElements(),
                                 criterion=AverageEnergy(alpha=0.8),
                                 method=SpecificRationalization(
            rationalizers=[RemoveShortSide(ratio=5.0),
                           QuadSplit(angle=150),
                           RemoveBadTriangle(acute_angle=15,obtuse_angle=150)])
                                 )
            )
    OOF.Skeleton.PinNodes.Pin_Internal_Boundary_Nodes(skeleton=skelname)
    OOF.Skeleton.Modify(
        skeleton=skelname,
        modifier=Smooth(targets=AllNodes(),
                        criterion=AverageEnergy(alpha=0.3),
                        T=0.0,
                        iteration=FixedIteration(iterations=5)))
    OOF.Skeleton.PinNodes.Undo(skeleton=skelname)


skeletonmenu.skeletonmenu.addItem(oofmenu.OOFMenuItem(
    'Auto',
    callback=_autoSkeleton,
    threadable=oofmenu.THREADABLE,
    no_log=True,                        # because it calls other menu items
    params=parameter.ParameterGroup(
    parameter.AutomaticNameParameter('name',
                           value=automatic.automatic,
                           resolver=skeletonmenu.skeletonNameResolver,
                           tip="Name of the new skeleton."),
    whoville.WhoParameter('microstructure',
                          oof2.common.microstructure.microStructures,
                          tip=parameter.emptyTipString),
    parameter.BooleanParameter('left_right_periodicity', value=False,
     tip="Whether or not the skeleton is periodic in the horizontal direction"),
    parameter.BooleanParameter("top_bottom_periodicity", value=False,
       tip="Whether or not the skeleton is periodic in the vertical direction"),
    parameter.FloatParameter('maxscale', value=1.0,
                             tip="Rough size of the largest elements."),
    parameter.FloatParameter("minscale", value=1.0,
                             tip="Rough size of the smallest elements."),
    enum.EnumParameter('units', oof2.common.units.Units, value='Physical',
                       tip="Units for minscale and maxscale."),
    parameter.FloatRangeParameter('threshold', (0.0, 1.0, 0.01),
                                  value=0.90,
                                  tip="Minimum acceptable homogeneity")
    ),                                  # end of ParameterGroup
    help="Automatically create and refine a &skel;.",
    discussion=xmlmenudump.loadFile("DISCUSSIONS/engine/menu/autoskel.xml")
    ))
