// -*- C++ -*-
// $RCSfile: femesh.C,v $
// $Revision: 1.165 $
// $Author: rlua $
// $Date: 2007/03/07 01:27:24 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include <iostream>
#include <string>
#include <math.h>
#include <vector>

#include "common/lock.h"
#include "engine/cscpatch.h"
#include "engine/csrmatrix.h"
#include "engine/csubproblem.h"
#include "engine/element.h"
#include "engine/elementnodeiterator.h"
#include "engine/equation.h"
#include "engine/flux.h"
#include "engine/field.h"
#include "engine/material.h"
#include "engine/femesh.h"
#include "engine/gausspoint.h"
#include "engine/mastercoord.h"
#include "engine/meshiterator.h"
#include "engine/nodalequation.h"
#include "engine/nodalfluxes.h"
#include "engine/nodalscpatches.h"
#include "engine/node.h"
#include "engine/ooferror.h"
#include "engine/outputval.h"
#include "engine/preconditioner.h"
#include "engine/property.h"
#include "engine/smallmatrix.h"
#include "engine/solver.h"
#include "engine/sparselink.h"
#include "common/tostring.h"
#include "common/trace.h"
#include "common/printvec.h"
#include "common/switchboard.h"
#include "common/IO/progressbar.h"

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

FEMesh::FEMesh(CMicrostructure * mc)
  : microstructure(mc),
    rwlock(0),
    ncount(0),
    dof_list_needs_cleaning(false),
    nodaleqn_list_needs_cleaning(false),
    defaultSubProblem(0)
{
}

FEMesh::~FEMesh() {
  for(std::vector<NodalEquation*>::size_type i=0; i<nodaleqn.size(); i++)
    delete nodaleqn[i];
  for(std::vector<DegreeOfFreedom*>::size_type i=0; i<dof.size(); i++)
    delete dof[i];
  for(std::vector<FuncNode*>::size_type i=0; i<funcnode.size(); ++i)
    delete funcnode[i];
  for(std::vector<Node*>::size_type i=0; i<mapnode.size(); ++i)
    delete mapnode[i];
  for(std::vector<Element*>::size_type i=0; i<element.size(); ++i)
    delete element[i];
  // We do not own rwlock, so don't delete it.

  std::map<const int, NodalSCPatches*>::iterator piter;
  for(piter=scpatches.begin(); piter!=scpatches.end(); piter++)
    delete piter->second;

  std::map<const int, NodalFluxes*>::iterator fiter;
  for(fiter=recovered_fluxes.begin(); fiter!=recovered_fluxes.end(); fiter++)
    delete fiter->second;


#ifdef HAVE_MPI
  for(std::map<FuncNode*,CNodeShareInfo*>::iterator it=m_nodesharemap.begin();
      it!=m_nodesharemap.end();it++)
    {
      delete it->second;
    }
  m_nodesharemap.clear();
#endif
}

const std::string &FEMesh::classname() const {
  static const std::string _name = "FEMesh";
  return _name;
}

const std::string &FEMesh::modulename() const {
  static const std::string _name = "oof2.SWIG.engine.femesh";
  return _name;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// All Degrees of Freedom are created and deleted with these functions

DegreeOfFreedom *FEMesh::CreateDoF(double x) {
  int index = dof.size();
  DegreeOfFreedom *newdof = new DegreeOfFreedom(x, index);
  dof.push_back(newdof);
  return newdof;
}

void FEMesh::removeDoF(DegreeOfFreedom *deaddof) {
//   Trace("FEMesh::removeDoF");
  int index = deaddof->dofindex();
  delete deaddof;
  // Don't shift the rest of the vector entries to fill the hole
  // caused by deleting this dof.  The hole can be filled more
  // efficiently by garbage collection later.  Just make sure to check
  // for null pointers when using the dof list!
  dof[index] = 0;
  dof_list_needs_cleaning = true;
}

void FEMesh::clean_doflist() {
  if(dof_list_needs_cleaning) {
    std::vector<DegreeOfFreedom*>::size_type j=0;
    for(std::vector<DegreeOfFreedom*>::size_type i=0; i<dof.size(); i++) {
      if (dof[i] != 0) {
	if(i != j) {
	  dof[j] = dof[i];
	  dof[j]->index_ = j;
	}
	j++;
      }
    }
    if(j != dof.size())
      dof.resize(j);
    dof_list_needs_cleaning=false;
  }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// All NodalEquations are created and deleted with these functions

NodalEquation *FEMesh::CreateNodalEqn() {
//   Trace("FEMesh::CreateNodalEqn");
  int index = nodaleqn.size();
  NodalEquation *neweqn = new NodalEquation(index);
  nodaleqn.push_back(neweqn);
  return neweqn;
}

void FEMesh::removeNodalEqn(NodalEquation *eqn) {
//   Trace("FEMesh::removeNodalEqn");
  int index = eqn->ndq_index();
  delete eqn;
  // Don't shift the rest of the vector entries to fill the hole
  // caused by deleting this eqn.  The hole can be filled more
  // efficiently by garbage collection later.  Just make sure to check
  // for null pointers when using the dof list!
  nodaleqn[index] = 0;
  nodaleqn_list_needs_cleaning=true;
}

void FEMesh::clean_nodaleqn() {
  if(nodaleqn_list_needs_cleaning) {
    std::vector<NodalEquation*>::size_type j = 0;
    for(std::vector<NodalEquation*>::size_type i=0; i<nodaleqn.size(); i++) {
      if(nodaleqn[i] != 0) {
	if(i != j) {
	  nodaleqn[j] = nodaleqn[i];
	  nodaleqn[j]->index_ = j;
	}
	j++;
      }
    }
    if(j != nodaleqn.size())
      nodaleqn.resize(j);
    nodaleqn_list_needs_cleaning = false;
  }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Make sure that all dofs and nodalequations have the right indices

void FEMesh::housekeeping() {
  clean_doflist();
  clean_nodaleqn();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

Node *FEMesh::newMapNode(const Coord &pos) {
  Node *node = new Node(ncount++, pos);
  mapnode.push_back(node);
  return node;
}

FuncNode *FEMesh::newFuncNode(const Coord &pos) {
  FuncNode *node = new FuncNode(this, ncount++, pos);
  funcnode.push_back(node);
  return node;
}

#ifdef HAVE_MPI

#include "common/mpitools.h"

//procs is (pointer to) an array of process numbers that share the node.
//indices is an array of (remote) indices of the node relative to other processes.
//procs and indices are 1-1 and must have the same length.
FuncNode *FEMesh::newFuncNode_shares(const Coord &pos, std::vector<int>* procs,
				     std::vector<int>* indices, int index)
{
  FuncNode *node = new FuncNode(this, ncount++, pos);
  funcnode.push_back(node);

  if(procs->size()>0 && procs->size()==indices->size())
    {
      CNodeShareInfo* cn=new CNodeShareInfo(procs, indices, index);
      node->setShared();
      m_nodesharemap[node]=cn;
      m_indexnodemap[index]=node;
    }

  return node;
}

CNodeShareInfo::CNodeShareInfo(std::vector<int>* procs,
		 std::vector<int>* indices, int index)
{
  inheritedindex=index;
  localprocrank=Rank();//from mpitools.h
//   int minrank=localprocrank;
  for(std::vector<int>::size_type i=0;i<procs->size();i++)
    {
      int pi=(*procs)[i];
      remoteproclist.push_back(pi);
//       if(localprocrank>pi)
// 	minrank=pi;
      remoteindexlist.push_back((*indices)[i]);
    }

  //bubble sort remoteproclist (make corresponding change to remoteindexlist)
  int tmp,rsize=remoteproclist.size();
  for(std::vector<int>::size_type i=0;i<rsize;i++)
    {
      for(std::vector<int>::size_type j=i+1;j<rsize;j++)
	{
	  if(remoteproclist[i]>remoteproclist[j])
	    {
	      tmp=remoteproclist[i];
	      remoteproclist[i]=remoteproclist[j];
	      remoteproclist[j]=tmp;
	      tmp=remoteindexlist[i];
	      remoteindexlist[i]=remoteindexlist[j];
	      remoteindexlist[j]=tmp;
	    }
	}
    }
  // Find next higher rank
  nextrank=-1;
  nextindex=-1;
  for(std::vector<int>::size_type i=0;i<remoteproclist.size();i++)
    {
      if(localprocrank<remoteproclist[i])
	{
	  nextrank=remoteproclist[i];
	  nextindex=remoteindexlist[i];
	  break;
	}
    }

  if(localprocrank<=remoteproclist[0])
    {
      _owns0=true;
    }
  else
    {
      _owns0=false;
    }
  owns=_owns0;
};

CNodeShareInfo::~CNodeShareInfo()
{
  //Not necessary
  remoteproclist.clear();
  remoteindexlist.clear();
  localdofindexlist.clear();
  localeqnindexlist.clear();
  symmatrixdofindexlist.clear();
  symmatrixeqnindexlist.clear();
};

#endif

void FEMesh::reserveFuncNodes(int n) {
  funcnode.reserve(n);
}

void FEMesh::reserveMapNodes(int n) {
  mapnode.reserve(n);
}

void FEMesh::reserveElements(int n) {
  element.reserve(n);
}

int FEMesh::nnodes() const {
  // don't just return ncount here, since the node index is just an
  // identifier.  When nodes are deleted, ncount won't be decreased.
  return funcnode.size() + mapnode.size();
}

void FEMesh::AddElement(Element *el) {
  element.push_back(el);
  el->set_index(element.size()-1);
}

Element *FEMesh::getElement(int i) const {
  return element[i];
}

int FEMesh::nelements() const {
  return element.size();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void FEMesh::refreshMaterials(PyObject *skeleton) {
  for(ElementIterator ei=element_iterator(); !ei.end(); ++ei)
    ei.element()->refreshMaterial(skeleton);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

ElementIterator FEMesh::element_iterator() const {
  return ElementIterator(new MeshElementIterator(this));
}

NodeIterator FEMesh::node_iterator() const {
  return NodeIterator(new MeshNodeIterator(this));
}

FuncNodeIterator FEMesh::funcnode_iterator() const {
  return FuncNodeIterator(new MeshFuncNodeIterator(this));
}

// operator[] is deprecated...
// FuncNode *FEMesh::operator[](const FuncNodeIterator &ni) const {
//   return funcnode[ni.index];
// }

// Node *FEMesh::operator[](const NodeIterator &ni) const {
//   return getNode(ni.index);
// }

// Caution: NodeIterator::index is not necessarily the same as
// node.index().  The argument to FEMesh::getNode is the
// NodeIterator::index.

Node *FEMesh::getNode(int i) const {
  if(i < int(funcnode.size()))
    return funcnode[i];
  return mapnode[i - funcnode.size()];
}

FuncNode *FEMesh::getFuncNode(int i) const {
  return funcnode[i];
}

// Finding the closest node to the mouse point.
// Used in MeshInfo
// TODO LATER: use hash table lookup here, instead of looping over all nodes.
Node *FEMesh::closestNode(const double x, const double y) {
  double min=1.;   // (initial value provided to suppress compiler warnings)
  Node *node, *thenode=0;
  for(NodeIterator ni = node_iterator(); !ni.end(); ++ni) {
    node = ni.node();
    double dx = node->position()(0) - x;
    double dy = node->position()(1) - y;
    double dist = dx*dx + dy*dy;
    if (ni.begin()) {
      min = dist;
      thenode = node;
    }
    else {
      if (dist <= min) {
	min = dist;
	thenode = node;
      }
    }
  }
  return thenode;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void FEMesh::get_unknowns(VECTOR_D &x, const std::vector<int> &map) const {
  for(std::vector<DegreeOfFreedom*>::size_type i=0; i<dof.size(); i++) {
    if(map[i] != -1) {
      x[map[i]] += dof[i]->value();
    }
  }
}

void FEMesh::set_unknowns(const VECTOR_D &x, const std::vector<int> &map) {
//   Trace("FEMesh::set_unknowns");
//   dump_dof();
  for(std::vector<DegreeOfFreedom*>::size_type i=0; i<dof.size(); i++) {
    if(map[i] != -1) {
      //std::cerr << "Setting dof[" << i << "] = " << x[map[i]] << std::endl;
      dof[i]->value() = x[map[i]];
    }
  }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void FEMesh::set_in_plane(const Field &field, bool planarity) {
  if(field.index() >= in_plane_field.size())
    in_plane_field.resize(field.index()+1);
  in_plane_field[field.index()] = planarity;
}

bool FEMesh::in_plane(const Field &field) const {
  return (field.index() < in_plane_field.size()
	  && in_plane_field[field.index()]);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Return a list of Field names, in the order in which they were
// defined, for the FieldSet with the given id.

std::vector<std::string> *FEMesh::getFieldSetByID(int id) const {
  // Loop over all FEWrappers to find the one with this id. 
  for(FEWrapper<Field>::AllWrappers::const_iterator i=fieldwrappers.begin();
      i!=fieldwrappers.end(); ++i)
    {
      if((*i).second->id() == id) {
	FEWrapper<Field> &wrapper = *(*i).second;
	FEWrapper<Field>::FEvector &datalist = (*i).second->datalist;
	std::vector<std::string> *fieldnames = 
	  new std::vector<std::string>(wrapper.size);
	for(FEWrapper<Field>::FEvector::size_type f=0; f<datalist.size(); f++) {
	  if(datalist[f].order >= 0) {
	    (*fieldnames)[datalist[f].order] = getFieldByIndex(f)->name();
	  }
	}
	return fieldnames;
      }
    }
  throw ErrProgrammingError("Couldn't find FieldSet", __FILE__, __LINE__);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// ZZ error estimation stuff.  TODO: This should be encasulated
// elsewhere, so that different error estimators can be used.

void FEMesh::setDefaultSubProblem(CSubProblem *subp) {
  defaultSubProblem = subp;
}

// initializing "scpatches"
void FEMesh::init_scpatches(const std::vector<int> *snodes) {
  // snodes is a list of mesh node indices, in order of the
  // corresponding skeleton node indices.  scpatches is a std::map,
  // mapping ints to NodalSCPatches*.
  for(std::vector<int>::size_type i=0; i<snodes->size(); i++)
    scpatches[(*snodes)[i]] = new NodalSCPatches();
}

// create/add a new CSCPatch pointer
void FEMesh::add_scpatch(const int assembly_node, const Material *mat,
			 const int order,
			 const std::vector<int> *elems,
			 const std::vector<int> *nodes,
			 const int qualified)
{
  // make sure to delete new'd CSCPatch's at FEMesh's destruction.
  CSCPatch *patch = new CSCPatch(this, order, mat, elems, nodes, qualified);
  scpatches[assembly_node]->add_patch(patch);
}

// get elements and nodes from the patch
std::vector<int> *FEMesh::get_elements_from_patch(const int assembly_node,
						  const Material *mat)
{
  return scpatches[assembly_node]->get_elements_from_patch(mat);
}

std::vector<int> *FEMesh::get_nodes_from_patch(const int assembly_node,
					       const Material *mat)
{
  return scpatches[assembly_node]->get_nodes_from_patch(mat);
}

// initialize "recovered_fluxes"
void FEMesh::init_nodalfluxes() {
  for(FuncNodeIterator i=funcnode_iterator(); !i.end(); ++i)
    recovered_fluxes[i.node()->index()] = new NodalFluxes();
}

// recover fluxes
void FEMesh::recover_fluxes() {
  std::map<const int, NodalSCPatches*>::iterator iter;
  std::vector<Flux*> allfluxes = defaultSubProblem->allFluxes();
  for(iter=scpatches.begin(); iter!=scpatches.end(); iter++)
    (iter->second)->recover_fluxes(allfluxes);
}

// add a recovered flux to the map
void FEMesh::add_this_flux(const Material *mat, const Flux *fluks,
			   const Node *nd, VECTOR_D *fvalues)
{
  recovered_fluxes[nd->index()]->add_flux_value(mat, fluks, fvalues);
}

// compute a recovered flux at MasterCoord
VECTOR_D *FEMesh::get_recovered_flux(const Flux *fluks, const Element *elem,
				     const MasterCoord &mc)
{
  int dim = fluks->ndof();
  VECTOR_D *result = new VECTOR_D(dim, 0.0);
  for(int i=0; i<dim; i++) {
    // interpolating the value
    for(ElementFuncNodeIterator node = elem->funcnode_iterator(); !node.end();
	++node)
      {
	double temp = recovered_fluxes[(node.node())->index()]->
	  get_flux_component(elem->material(), fluks, i);
	(*result)(i) += node.shapefunction(mc)*temp;
      }
  }
  return result;
}


// compute all the recovered fluxes at Coord -- swigged one
void FEMesh::report_recovered_fluxes(const Element *elem, const Coord *point) {
  std::vector<Flux*> fluxes = defaultSubProblem->allFluxes();
  for (std::vector<Flux*>::size_type i=0; i<fluxes.size(); i++) {
    std::cerr << fluxes[i]->name() << std::endl;
    int dim = fluxes[i]->ndof();
    for(int j=0; j<dim; j++) {
      double j_th_component = 0.0;
      for(ElementFuncNodeIterator node = elem->funcnode_iterator(); !node.end();
	  ++node)
	{
	  double comp = recovered_fluxes[(node.node())->index()]
	    ->get_flux_component(elem->material(), fluxes[i], j);
	  j_th_component += node.shapefunction(elem->to_master(*point))*comp;
	}
      std::cerr << j_th_component << std::endl;
    }
    std::cerr << "==========================" << std::endl;
  }
}


double FEMesh::zz_L2_estimate(const Element *elem, const Flux *fluks) {
  // What to compute:
  // Integral_e[(flux_ex-flux_fe)^T (flux_ex-flux_fe)]^(1/2)
  // flux_ex -> recovered flux
  // 
  // Things to consider to determine integration order
  // Linear mesh: recovered stress field is linear, so the order
  // is 2 -- 2x2 gauss point needed.
  // Quadratic mesh: recovered stress field is quadratic, thus the order
  // should be 4 -- 3x3 gauss points needed.

  int order = 2*elem->shapefun_degree();
  int dim = fluks->ndof();
  double error = 0.0;
  double refer = 0.0;
  for(GaussPointIterator gpt = elem->integrator(order); !gpt.end(); ++gpt) {
    MasterCoord mc = gpt.gausspoint().mastercoord();
    double wt = gpt.gausspoint().weight();
    zz_L2_estimate_sub(elem, fluks, dim, error, refer, mc, wt);
  }
  if (refer == 0.0)
    return 0.0;
  return sqrt(error/refer);
}


void FEMesh::zz_L2_estimate_sub(const Element *elem, const Flux *fluks,
				const int &dim, double &error, double &refer,
				const MasterCoord &mc, const double &wt)
{
  // flux(diff) vectors
  VECTOR_D *recovered = get_recovered_flux(fluks, elem, mc);
  VECTOR_D *feflux = fluks->evaluate(this, elem, mc);
  VECTOR_D diff = *recovered - *feflux;
  // dot product while integrating
  for(int i=0; i<dim; i++) {
    error += diff(i)*diff(i);
    refer += (*recovered)(i) * (*recovered)(i);
  }
  error *= wt;
  refer *= wt;
  delete recovered;
  delete feflux;
}


std::vector<double> *FEMesh::zz_L2_weights(const Flux *fluks,
					   const double &bottom,
					   const double &top)
{
  std::vector<double> *weights = new std::vector<double>;
  std::vector<double> values;
  
  int dim = fluks->ndof();
  bool first = true;
  int order=0;			// initialized to suppress compiler warnings
  double min=0, max=0;		// will be reinitialized below
  for(ElementIterator i=element_iterator(); !i.end(); ++i) {
    Element *elem = i.element();
    if(first)
      order = 2*elem->shapefun_degree();
    double value = 0.0;
    for(GaussPointIterator gpt = elem->integrator(order); !gpt.end(); ++gpt) {
      MasterCoord mc = gpt.gausspoint().mastercoord();
      double wt = gpt.gausspoint().weight();
      zz_L2_weights_sub(elem, fluks, dim, value, mc, wt);
    }
    values.push_back(value);
    if(first) {
      min = value;
      max = value;
      first = false;
    }
    else {
      if(value < min)
	min = value;
      if(value > max)
	max = value;
    }
  }

//   int order = 2*mesh->element[0]->shapefun_degree();
//   int dim = fluks->ndof();

//   Element *elem = element[0];
//   double value = 0.0;
//   for(GaussPointIterator gpt = elem->integrator(order); !gpt.end(); ++gpt) {
//     MasterCoord mc = gpt.gausspoint().mastercoord();
//     double wt = gpt.gausspoint().weight();
//     zz_L2_weights_sub(elem, fluks, dim, value, mc, wt);
//   }
//   values.push_back(value);
//   double min = value;
//   double max = value;

//   for(int i=1; i<mesh->nelements(); i++) {
//     Element *elem = mesh->element[i];
//     value = 0.0;
//     for(GaussPointIterator gpt = elem->integrator(order); !gpt.end(); ++gpt) {
//       MasterCoord mc = gpt.gausspoint().mastercoord();
//       double wt = gpt.gausspoint().weight();
//       zz_L2_weights_sub(elem, fluks, dim, value, mc, wt);
//     }
//     values.push_back(value);
//     if(value < min)
//       min = value;
//     if(value > max)
//       max = value;
//   }

  double upper = min + (max-min)*top;
  double lower = min + (max-min)*bottom;
  double diff = upper - lower;
  double sum = upper + lower;
  for(std::vector<double>::size_type i=0; i<values.size(); i++)
    weights->push_back(0.5*erf(M_E*(2*values[i]-sum)/diff)+0.5);
  return weights;
}

void FEMesh::zz_L2_weights_sub(const Element *elem, const Flux *fluks,
			       const int &dim, double &value,
			       const MasterCoord &mc, const double &wt)
{
  // flux vector
  VECTOR_D *feflux = fluks->evaluate(this, elem, mc);
  // dot product while integrating
  for(int i=0; i<dim; i++) {
    value += (*feflux)(i)*(*feflux)(i);
  }
  value *= wt;
  delete feflux;
}




